/***	obj -- object file routines.
 *	
 *	SYNOPSIS:
 *	  #include "obj.h"
 *	  uhdr = objnew (sIn);
 *	  usym = symfirst (sIn);
 *	  usym = symnext (sIn);
 *
 *	  TUHDR  *uhdr;
 *	  TUSYM  *usym;
 *	  FILE  *sIn;
 *
 *	DESCRIPTION:
 *	  These functions read headers and symbols from a.out, b.out, and x.out
 *	  format object files.  'Objnew' must always be called first; it reads
 *	  in the header, and decides what format things are in.  'Symfirst' and
 *	  'symnext' read in the first and subsequent symbols, respectively.
 *	  Thus a typical sequence would be:
 *	    ...
 *	    uhdr = objnew (sIn);
 *	    for (usym = symfirst (sIn); usym != NULL; usym = symnext (sIn))
 *	       whatever;
 *	    ...
 *
 *	RETURN VALUE:
 *	  See the individual function headers.
 *
 *	WARNINGS:
 *	  See the individual function headers.
 *
 *	EFFECTS:
 *	  See the individual function headers.
 *
 *	BUGS:
 *	  The return values point to static data.  Only one object file may
 *	  be processed at a time.
 *
 *	  The union used for symbols is arguably overly complex; perhaps there
 *	  should be separate routines asymfirst, bsymfirst, and xsymfirst, or
 *	  a single one returning a (char *) which can be cast by the user.
 *
 *	MODIFICATION HISTORY:
 *	M000	05 Dec 82	andyp
 *	- Created; much of code stolen from nlist(3), many thanks to vich
 */

# include <stdio.h>
# include <a.out.h>
# include <sys/relsym.h>
# include "obj.h"
# define EOS		'\0'

    /* NOT currently implemented, should be */
#ifdef M_WORDSWAP
#  define hostwswap 1
#else
#  define hostwswap 0
#endif

    /* routines return ptrs to these guys */
static TUHDR	uhdr;
static TUSYM	usym;

    /* state information kept in these guys btwn calls to routines */
static short	htype;		/* header type - values ("O*") in obj.h */
static long	lsyms;		/* # of bytes of syms not yet read */
static long	offbeg;		/* offset of header (beginning) of obj module */

    /* to stop from having to do xxx.yyy.zzz every time */
static struct xexec	*xhdrp	= &(uhdr.uh_hdr.uh_xhdr);
static struct aexec	*ahdrp	= &(uhdr.uh_hdr.uh_ahdr);
static struct bexec	*bhdrp	= &(uhdr.uh_hdr.uh_bhdr);
    /* ... */
static struct x2sym	*x2symp	= &(usym.us_sym.us_x2sym);
static struct asym	*asymp	= &(usym.us_sym.us_asym);
static struct b2sym	*b2symp	= &(usym.us_sym.us_b2sym);

/***	objnew -- initialize package for object module.
 *	
 *	SYNOPSIS:
 *	  uhdr = objnew (sIn)
 *	  TUHDR  *uhdr;
 *	  FILE  *sIn;
 *
 *	  uhdr:		Header of object file; NULL if unrecognized format.
 *	  sIn:		Input file stream.
 *
 *	DESCRIPTION:
 *
 *	RETURN VALUE:
 *	  Upon successful completion, a pointer to a TUHDR is returned.  Other-
 *	  wise, a value of NULL is returned.
 *
 *	WARNINGS:
 *	  The return value points to static data, whose contents are overwritten
 *	  on each call.
 *
 *	EFFECTS:
 *	  None.
 *
 *	BUGS:
 *
 *	MODIFICATION HISTORY:
 *	  Created	05 Dec 82	andyp
 */

TUHDR  *
objnew (sIn)

FILE  *sIn;
{
	extern long  ftell();
	extern TUHDR  uhdr;
	extern short  htype;
	extern struct xexec  *xhdrp;
	extern struct aexec  *ahdrp;
	extern struct bexec  *bhdrp;
	extern long  lsyms, offbeg;

	int  hsize;

	offbeg = ftell (sIn);
	hsize = fread (
		(char *) &(uhdr.uh_hdr),
		sizeof (char), sizeof (uhdr.uh_hdr),
		sIn
	);
	switch ((int) (bhdrp->xb_magic)) {
		case A_MAGIC1:
		case A_MAGIC2:
		case A_MAGIC3:
		case A_MAGIC4:
			if (hsize < sizeof (struct bexec)) {
				uhdr.uh_type = htype = ONOTOUT;
				return (NULL);
			}
			else {
# ifdef DEBUG
				dbprintf ("objnew: b.out\n");
# endif
				uhdr.uh_type = htype = OBOUT;
				lsyms = (bhdrp->xb_syms);
# ifdef DEBUG
				dbprintf ("objnew: lsyms: %d\n", (int) lsyms);
# endif
				return (&uhdr);
			}
		default:
			break;
	}
	switch (xhdrp->x_magic) {
		case X_MAGIC:
			if (hsize < sizeof (struct xexec)) {
				uhdr.uh_type = htype = ONOTOUT;
				return (NULL);
			}
			else {
# ifdef DEBUG
				dbprintf ("objnew: x.out\n");
# endif
				uhdr.uh_type = htype = OXOUT;
				lsyms = (long) (xhdrp->x_syms);
# ifdef DEBUG
				dbprintf ("objnew: lsyms: %d\n", (int) lsyms);
# endif
				return (&uhdr);
			}
		default:
			break;
	}
	switch (ahdrp->xa_magic) {
		case A_MAGIC1:
		case A_MAGIC2:
		case A_MAGIC3:
		case A_MAGIC4:
			if (hsize < sizeof (struct aexec)) {
				uhdr.uh_type = htype = ONOTOUT;
				return (NULL);
			}
			else {
# ifdef DEBUG
				dbprintf ("objnew: a.out\n");
# endif
				uhdr.uh_type = htype = OAOUT;
				lsyms = (ahdrp->xa_syms);
# ifdef DEBUG
				dbprintf ("objnew: lsyms: %d\n", (int) lsyms);
# endif
				return (&uhdr);
			}
		default:
			break;
	}
	uhdr.uh_type = htype = ONOTOUT;	/* so that call to symfirst will bomb */
	return (NULL);
}

/***	symfirst -- get first symbol from object module name list.
 *	
 *	SYNOPSIS:
 *	  usym = symfirst (sIn)
 *	  TUSYM  *usym;
 *	  FILE  *sIn;
 *
 *	  usym:		First symbol in name list; NULL if no symbols.
 *	  sIn:		Input file stream.
 *
 *	DESCRIPTION:
 *
 *	RETURN VALUE:
 *	  Upon successful completion, a pointer to a TUSYM is returned.  Other-
 *	  wise, a value of NULL is returned.
 *
 *	WARNINGS:
 *	  The return value points to static data, whose contents are overwritten
 *	  on each call.
 *
 *	EFFECTS:
 *	  None.
 *
 *	BUGS:
 *	  NULL is returned both for no symbols (not an error) and for errors.
 *
 *	MODIFICATION HISTORY:
 *	  Created	05 Dec 82	andyp
 */

TUSYM  *
symfirst (sIn)

FILE  *sIn;
{
	extern TUSYM *symnext();
	extern short  htype;

	switch (htype) {
		case OAOUT:
			fseek (sIn, offbeg + ASYMPOS (ahdrp), 0);
			return (symnext (sIn));
		case OBOUT:
			fseek (sIn, offbeg + BSYMPOS (bhdrp), 0);
			return (symnext (sIn));
		case OXOUT:
			fseek (sIn, offbeg + XSYMPOS (xhdrp), 0);
			return (symnext (sIn));
		case ONOTOUT:
		default:
			return (NULL);
			break;
	}
	return (NULL);
}

/***	symnext -- get next symbol from object module name list.
 *	
 *	SYNOPSIS:
 *	  usym = symnext (sIn)
 *	  TUSYM  *usym;
 *	  FILE  *sIn;
 *
 *	  usym:		Next symbol in name list; NULL if no symbols left.
 *	  sIn:		Input file stream.
 *
 *	DESCRIPTION:
 *
 *	RETURN VALUE:
 *	  Upon successful completion, a pointer to a TUSYM is returned.  Other-
 *	  wise, a value of NULL is returned.
 *
 *	WARNINGS:
 *	  The return value points to static data, whose contents are overwritten
 *	  on each call.
 *
 *	EFFECTS:
 *	  None.
 *
 *	BUGS:
 *	  NULL is returned both for no more symbols (not an error) and for
 *	  errors.
 *
 *	MODIFICATION HISTORY:
 *	  Created	05 Dec 82	andyp
 */

TUSYM  *
symnext (sIn)

FILE  *sIn;
{
	extern int  strnget();
	extern short  htype;
	extern long  lsyms;
	extern struct x2sym  *x2symp;
	extern struct asym  *asymp;
	extern struct b2sym  *b2symp;
	int  n, ssize;

	switch (htype) {
		case OAOUT:
			if (lsyms <= 0) {
# ifdef DEBUG
				dbprintf ("symnext: lsyms <= 0\n");
# endif
				return (NULL);
			}
			ssize = fread (
				(char *) asymp,
				sizeof (char), sizeof (struct asym),
				sIn
			);
			if (ssize != sizeof (struct asym)) {
# ifdef DEBUG
				dbprintf ("symnext: EOF, didn't find sym\n");
# endif
				return (NULL);
			}
			usym.us_type = OAOUT;
			lsyms -= (long) ssize;
# ifdef DEBUG
			dbprintf ("symnext: got one\n");
# endif
			return (&usym);
		case OBOUT:
			if (lsyms <= 0) {
# ifdef DEBUG
				dbprintf ("symnext: lsyms <= 0\n");
# endif
				return (NULL);
			}
			/*
			 * n.b. we read 'sizeof (struct bsym)' (not b2sym) into
			 * 'struct b2sym'.  Then we read in the symbol name.
			 */
			ssize = fread (
				(char *) b2symp,
				sizeof (char), sizeof (struct bsym),
				sIn
			);
			if (ssize != sizeof (struct bsym)) {
# ifdef DEBUG
				dbprintf ("symnext: EOF, didn't find sym\n");
# endif
				return (NULL);
			}
			/* variable length name follows symbol */
			n = strnget (sIn, &(b2symp->sb2_name [0]), SYMLENGTH);
			if (n <= 0) {
# ifdef DEBUG
				dbprintf ("symnext: strnget returns neg\n");
# endif
				return (NULL);
			}
			usym.us_type = OBOUT;
			lsyms -= (long) ssize + (long) n;
# ifdef DEBUG
			dbprintf ("symnext: got one\n");
# endif
			return (&usym);
		case OXOUT:
			if (lsyms <= 0) {
# ifdef DEBUG
				dbprintf ("symnext: lsyms <= 0\n");
# endif
				return (NULL);
			}
			/*
			 * n.b. we read 'sizeof (struct sym)' (not x2sym) into
			 * 'struct x2sym'.  Then we read in the symbol name.
			 */
			ssize = fread (
				(char *) x2symp,
				sizeof (char), sizeof (struct sym),
				sIn
			);
			if (ssize != sizeof (struct sym)) {
# ifdef DEBUG
				dbprintf ("symnext: EOF, didn't find sym\n");
# endif
				return (NULL);
			}
			/* variable length name follows symbol */
			n = strnget (sIn, &(x2symp->sx2_name [0]), SYMLENGTH);
			if (n <= 0) {
# ifdef DEBUG
				dbprintf ("symnext: strnget returns neg\n");
# endif
				return (NULL);
			}
			usym.us_type = OXOUT;
			lsyms -= (long) ssize + (long) n;
# ifdef DEBUG
			dbprintf ("symnext: got one\n");
# endif
			return (&usym);
		default:
		case ONOTOUT:
			return (NULL);
			break;
	}
	return (NULL);
}

static int
strnget (sIn, name, n)

FILE  *sIn;
char  *name;
int  n;
{
	char  ch;
	int  c, i;

	i = 0;
	do {
		c = getc (sIn);
		if (c == EOF || i++ >= n)
			return (0);
		*name++ = ch = (char) c;
	} while (ch != EOS);
	return (i);
}

/*VARARGS1*/
static
dbprintf (fmt, a1, a2, a3, a4, a5)

char  *fmt;
int  a1, a2, a3, a4, a5;
{
	printf ("debug: ");
	printf (fmt, a1, a2, a3, a4, a5);
}
